/*
 * PreZ v1.0 - FreeBSD Version
 * Allows you to start a new thread at any program.
 * The program consists of three parts, the injector, the thread creator
 * and the code that will be executed. This is the first part. Check
 * inject.s for the second and third. After compiling inject.s you must
 * create the code.h file which has all code stored in the injcode variable.
 * You can create this file using objdump and the provided makecode.pl program.
 *
 * By Fotis Loukos <fotisl@gmail.com>
 *
 * Caer está permitido. Levantarse es obligatorio.
 * - Pablo Garcia
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include <machine/reg.h>
#include "code.h"

int main(int argc, char **argv)
{
    pid_t proc;
    struct reg regs;
    struct ptrace_io_desc iodesc;
    char *oldcode;
    int len;

    if(argc != 2) {
        printf("Usage: %s <pid>\n", argv[0]);
        exit(1);
    }

    proc = (pid_t) atoi(argv[1]);

    len = sizeof(injcode);
    if((oldcode = malloc(len)) == NULL) {
        perror("Cannot allocate memory for old code");
        exit(1);
    }

    printf("Attaching to process...\n");
    if(ptrace(PT_ATTACH, proc, NULL, 0) < 0) {
        perror("Cannot attach to process");
        exit(1);
    }
    waitpid(proc, NULL, 0);

    printf("Attached! Getting register info...\n");
    ptrace(PT_GETREGS, proc, (caddr_t) &regs, 0);

    printf("Stopped execution at 0x%08x. Overwritting code...\n", regs.r_eip);
    iodesc.piod_op = PIOD_READ_I;
    iodesc.piod_offs = (void *) regs.r_eip;
    iodesc.piod_addr = oldcode;
    iodesc.piod_len = len;
    ptrace(PT_IO, proc, (caddr_t) &iodesc, 0);
    iodesc.piod_op = PIOD_WRITE_I;
    iodesc.piod_offs = (void *) regs.r_eip;
    iodesc.piod_addr = injcode;
    iodesc.piod_len = len;
    ptrace(PT_IO, proc, (caddr_t) &iodesc, 0);

    printf("Resuming operation...\n");
    ptrace(PT_CONTINUE, proc, (caddr_t) 1, 0);

    waitpid(proc, NULL, 0);

    printf("Code injected and running, restoring old code and registers...\n");
    iodesc.piod_op = PIOD_WRITE_I;
    iodesc.piod_offs = (void *) regs.r_eip;
    iodesc.piod_addr = oldcode;
    iodesc.piod_len = len;
    ptrace(PT_IO, proc, (caddr_t) &iodesc, 0);
    ptrace(PT_SETREGS, proc, (caddr_t) &regs, 0);

    printf("Detaching...\n");
    ptrace(PT_DETACH, proc, (caddr_t) 1, 0);

    free(oldcode);

    return 0;
}
