/*
 * PreZ v1.0 - Linux Version
 * Allows you to start a new thread at any program.
 * The program consists of three parts, the injector, the thread creator
 * and the code that will be executed. This is the first part. Check
 * inject.s for the second and third. After compiling inject.s you must
 * create the code.h file which has all code stored in the injcode variable.
 * You can create this file using objdump and the provided makecode.pl program.
 *
 * By Fotis Loukos <fotisl@gmail.com>
 *
 * Caer está permitido. Levantarse es obligatorio.
 * - Pablo Garcia
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/user.h>
#include <sys/ptrace.h>
#include "code.h"

void readmem(pid_t proc, unsigned long addr, char *buffer, int length)
{
    long *p;
    int i, j;

    j = length / sizeof(long);
    p = (long *) buffer;
    for(i = 0; i < j; i++)
        *p++ = ptrace(PTRACE_PEEKDATA, proc, addr + i * sizeof(long), NULL);

    if((length % sizeof(long)) != 0) {
        p = (long *) &buffer[length - 4];
        *p = ptrace(PTRACE_PEEKDATA, proc, addr + length - 4, NULL);
    }
}

void writemem(pid_t proc, unsigned long addr, char *buffer, int length)
{
    long *p;
    int i, j;

    j = length / sizeof(long);
    p = (long *) buffer;
    for(i = 0; i < j; i++)
        ptrace(PTRACE_POKEDATA, proc, addr + i * sizeof(long), *p++);

    if((length % sizeof(long)) != 0) {
        p = (long *) &buffer[length - 4];
        ptrace(PTRACE_POKEDATA, proc, addr + length - 4, *p);
    }
}

int main(int argc, char **argv)
{
    pid_t proc;
    struct user_regs_struct regs;
    char *oldcode;
    int len;

    if(argc != 2) {
        printf("Usage: %s <pid>\n", argv[0]);
        exit(1);
    }

    proc = (pid_t) atoi(argv[1]);

    len = sizeof(injcode);
    if((oldcode = malloc(len)) == NULL) {
        perror("Cannot allocate memory for old code");
        exit(1);
    }

    printf("Attaching to process...\n");
    if(ptrace(PTRACE_ATTACH, proc, NULL, NULL) < 0) {
        perror("Cannot attach to process");
        exit(1);
    }
    waitpid(proc, NULL, 0);

    printf("Attached! Getting register info...\n");
    ptrace(PTRACE_GETREGS, proc, NULL, &regs);

    printf("Stopped execution at 0x%08lx. Overwritting code...\n", regs.eip);
    readmem(proc, regs.eip, oldcode, len);
    writemem(proc, regs.eip, (char *) injcode, len);

    printf("Resuming operation...\n");
    ptrace(PTRACE_CONT, proc, NULL, NULL);

    waitpid(proc, NULL, 0);

    printf("Code injected and running, restoring old code and registers...\n");
    writemem(proc, regs.eip, oldcode, len);
    ptrace(PTRACE_SETREGS, proc, NULL, &regs);

    printf("Detaching...\n");
    ptrace(PTRACE_DETACH, proc, NULL, NULL);

    free(oldcode);

    return 0;
}
