/*
 * PreZ v1.0 - OpenSolaris Version
 * Allows you to start a new thread at any program.
 * The program consists of three parts, the injector, the thread creator
 * and the code that will be executed. This is the first part. Check
 * inject.s for the second and third. After compiling inject.s you must
 * create the code.h file which has all code stored in the injcode variable.
 * You can create this file using objdump and the provided makecode.pl program.
 *
 * By Fotis Loukos <fotisl@gmail.com>
 *
 * Caer está permitido. Levantarse es obligatorio.
 * - Pablo Garcia
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <procfs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/regset.h>
#include <sys/fault.h>
#include "code.h"

int main(int argc, char **argv)
{
    int pid, size;
    int ctlfd, statusfd, lwpstatusfd, lwpctlfd, asfd;
    long *ctlmsg;
    char buf[128];
    unsigned char *oldcode;
    pstatus_t pstatus;
    lwpstatus_t lwpstatus;
    fltset_t fltset;

    if(argc != 2) {
        printf("Usage: %s <pid>\n", argv[0]);
        exit(1);
    }

    pid = atoi(argv[1]);

    if((oldcode = malloc(sizeof(injcode))) == NULL) {
        perror("Cannot allocate memory for old code");
        exit(1);
    }

    size = ((sizeof(prgregset_t) > sizeof(fltset_t)) ? sizeof(prgregset_t) :
        sizeof(fltset_t)) + sizeof(long);
    if((ctlmsg = malloc(size)) == NULL) {
        perror("Cannot allocate memory for control messages");
        free(oldcode);
        exit(1);
    }

    snprintf(buf, 128, "/proc/%i/ctl", pid);
    printf("Opening ctl file (%s)...\n", buf);
    if((ctlfd = open(buf, O_WRONLY)) < 0) {
        perror("Cannot open ctl file");
        free(oldcode);
        free(ctlmsg);
        exit(1);
    }

    snprintf(buf, 128, "/proc/%i/status", pid);
    printf("Opening status file (%s)...\n", buf);
    if((statusfd = open(buf, O_RDONLY)) < 0) {
        perror("Cannot open status file");
        close(ctlfd);
        free(oldcode);
        free(ctlmsg);
        exit(1);
    }

    snprintf(buf, 128, "/proc/%i/as", pid);
    printf("Opening as file (%s)...\n", buf);
    if((asfd = open(buf, O_RDWR)) < 0) {
        perror("Cannot open as file");
        close(statusfd);
        close(ctlfd);
        free(oldcode);
        free(ctlmsg);
        exit(1);
    }

    snprintf(buf, 128, "/proc/%i/lwp/1/lwpctl", pid);
    printf("Opening thread 1 lwpctl file (%s)...\n", buf);
    if((lwpctlfd = open(buf, O_WRONLY)) < 0) {
        perror("Cannot open lwpctl file");
        close(asfd);
        close(statusfd);
        close(ctlfd);
        free(oldcode);
        free(ctlmsg);
        exit(1);
    }

    snprintf(buf, 128, "/proc/%i/lwp/1/lwpstatus", pid);
    printf("Opening thread 1 lwpstatus file (%s)...\n", buf);
    if((lwpstatusfd = open(buf, O_RDONLY)) < 0) {
        perror("Cannot open status file");
        close(lwpctlfd);
        close(asfd);
        close(statusfd);
        close(ctlfd);
        free(oldcode);
        free(ctlmsg);
        exit(1);
    }

    printf("Stopping program...\n");
    ctlmsg[0] = PCSTOP;
    write(ctlfd, ctlmsg, sizeof(long));

    printf("Program stopped, reading status...\n");

    read(lwpstatusfd, &lwpstatus, sizeof(lwpstatus));
    read(statusfd, &pstatus, sizeof(pstatus));

    printf("Program stopped at 0x%08x.\n", lwpstatus.pr_reg[EIP]);

    printf("Setting faults to be traced...");

    premptyset(&fltset);
    praddset(&fltset, FLTBPT);
    ctlmsg[0] = PCSFAULT;
    memcpy((void *) (ctlmsg + 1), (void *) &fltset, sizeof(fltset));
    write(ctlfd, ctlmsg, sizeof(long) + sizeof(fltset));

    printf("Injecting code...\n");

    lseek(asfd, lwpstatus.pr_reg[EIP], SEEK_SET);
    read(asfd, oldcode, sizeof(injcode));
    lseek(asfd, lwpstatus.pr_reg[EIP], SEEK_SET);
    write(asfd, injcode, sizeof(injcode));

    printf("Code injected, continuing thread...\n");

    ctlmsg[0] = PCRUN;
    ctlmsg[1] = 0;
    write(lwpctlfd, ctlmsg, 2 * sizeof(long));

    printf("Waiting for thread to stop by breakpoint...\n");

    ctlmsg[0] = PCWSTOP;
    write(ctlfd, ctlmsg, sizeof(long));

    printf("Clearing fault...\n");

    ctlmsg[0] = PCCFAULT;
    write(ctlfd, ctlmsg, sizeof(long));

    printf("Thread stopped, restoring old code, registers and traced faults...\n");

    lseek(asfd, lwpstatus.pr_reg[EIP], SEEK_SET);
    write(asfd, oldcode, sizeof(injcode));

    ctlmsg[0] = PCSREG;
    memcpy((void *) (ctlmsg + 1), (void *) lwpstatus.pr_reg, sizeof(lwpstatus.pr_reg));
    write(lwpctlfd, ctlmsg, sizeof(long) + sizeof(lwpstatus.pr_reg));

    ctlmsg[0] = PCSFAULT;
    memcpy((void *) (ctlmsg + 1), (void *) &pstatus.pr_flttrace, sizeof(pstatus.pr_flttrace));
    write(ctlfd, ctlmsg, sizeof(long) + sizeof(pstatus.pr_flttrace));

    printf("Restarting all threads...\n");
    ctlmsg[0] = PCRUN;
    ctlmsg[1] = 0;
    write(ctlfd, ctlmsg, 2 * sizeof(long));

    printf("Done!\n");

    close(lwpstatusfd);
    close(lwpctlfd);
    close(asfd);
    close(statusfd);
    close(ctlfd);
    free(oldcode);
    free(ctlmsg);

    return 0;
}
